/*
 * direntries.c
 * manage entries (not inodes!) within a directory
 * Frank Bergmann, 9/2008
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <errno.h>
#include <stdint.h>
#include <string.h>

#define PATHNAME_SIZE 16384

const char usage[]="usage:\n"
  "direntries [directory]\n"
  "direntries -r [directory] # recursive mode\n"
  "direntries -m [-d depth] [directory] # maximum mode\n"
  "direntries -s [-p] [-e file-extension] [-P filename-prefix] [directory] # show mode\n"
  "direntries -D [-e file-extension] [-P filename-prefix] [directory] # delete mode\n";

enum { MODE_NORMAL, MODE_RECURSIVE, MODE_MAXIMUM, MODE_SHOW, MODE_DELETE };

static inline void print(const char *str) {
  write(1,str,strlen(str));
}

void my_exit(const int retval) {
  _exit(retval);
}

int64_t str2uint63(register const char *s) {
  register int64_t i;
  register char c;

  c = *s++;
  if ((c<'0') || (c>'9')) return -1;
  i = c-'0';
  for(;;) {
    c=*s;
    if((c<'0')||(c>'9')) {
      return i;
    } else {
      i=(i*10)+c-'0';
      s++;
    }
  }
}

int str_catn(char *out,const char *in, unsigned long limit) {
  register char* s=out;
  register const char* t=in;
  register const char* u=t+limit;
  s+=strlen(s);
  for (;;) {
    if (t>=u) { *s='\0'; break; }
    if (!(*s=*t)) break;
    ++s; ++t;
  }
  return (int)(s-out);
}

int str_diffn(register const char* s, register const char* t, uint32_t limit) {
  register const char* u=t+limit;
  register int j;
  j=0;
  for (;;) {
    if (t>=u) break; if ((j=(*s-*t))) break; if (!*t) break; ++s; ++t;
    if (t>=u) break; if ((j=(*s-*t))) break; if (!*t) break; ++s; ++t;
    if (t>=u) break; if ((j=(*s-*t))) break; if (!*t) break; ++s; ++t;
    if (t>=u) break; if ((j=(*s-*t))) break; if (!*t) break; ++s; ++t;
  }
  return j;
}

int32_t str_index(const char* a, const char* b) {
  register const char* s=a;
  register int t=strlen(b);
  register const char* u=s+strlen(s)-t;
  register char z;

  z = *b;
  for (;;) {
    /* if (s>u) break; if (!str_diffn(s,b,t)) return s-a; ++s; */
    if (s>u) break; if ((*s==z) && (!str_diffn(s,b,t))) return s-a; ++s;
    if (s>u) break; if ((*s==z) && (!str_diffn(s,b,t))) return s-a; ++s;
    if (s>u) break; if ((*s==z) && (!str_diffn(s,b,t))) return s-a; ++s;
    if (s>u) break; if ((*s==z) && (!str_diffn(s,b,t))) return s-a; ++s;
  }
  return -1;
}

int str_diff_extension(const char *filename, const char *file_extension) {
  register const char *s;
  register const char *t;
  register int j;
  register int32_t m,n;

  m=strlen(filename);n=strlen(file_extension);
  if(m<=n) return -1;
  s=filename+m-n;
  t=file_extension;
  for(;;) {
    if((j=(*s-*t))) break; if (!*t) break; if (!*s) break; ++s; ++t;
  }
  return j;
}

int walkdir(
    const char *name,
    uint32_t *entries,
    uint32_t *entries_maxdir,
    char *pathname_maxdir,
    uint16_t pathname_size,
    int32_t depth,
    int32_t curdepth,
    const char *subdir) {

  register DIR *dptr;
  register struct dirent *eptr;
  struct stat statbuf;
  register uint64_t entries_sum;
  register char *nptr;
  register int dotdirs;
  register int32_t hlinks;
  register int ret;
  char pathname[PATHNAME_SIZE*3];

  dptr=opendir(name);
  if(dptr==NULL) return 1;
  ret=0;
  dotdirs=0;
  hlinks=2;
  if(lstat(name,&statbuf)==0) {
    hlinks=statbuf.st_nlink;
    if(hlinks>=2) hlinks-=2; /* leaf optimization */
  }
  for(entries_sum=1;;) {
    eptr=readdir(dptr); if(eptr==NULL) break;
    nptr=eptr->d_name;
    if((dotdirs<2)&&(*nptr=='.')) {
      if(nptr[1]=='\0') { dotdirs+=1;continue; }
      if(nptr[1]=='.') {
        if(nptr[2]=='\0') { dotdirs+=1;continue; }
      }
    }
    if(hlinks>0) { /* leaf optimization */
      *pathname='\0';
      str_catn(pathname,name,PATHNAME_SIZE);
      str_catn(pathname,"/",1);
      str_catn(pathname,nptr,PATHNAME_SIZE);
      if(subdir!=NULL) {
        str_catn(pathname,"/",1);
        str_catn(pathname,subdir,PATHNAME_SIZE);
      }
      if(lstat(pathname,&statbuf)==0){
        if((statbuf.st_mode&S_IFMT)==S_IFDIR) {
          hlinks--;
          if(depth>0) {
            if(curdepth<depth) {
              ret+=walkdir(pathname,entries,entries_maxdir,pathname_maxdir,pathname_size,depth,curdepth+1,NULL);
            }
          } else {
            ret+=walkdir(pathname,entries,entries_maxdir,pathname_maxdir,pathname_size,depth,0,NULL);
          } /* depth */
        } else {
          entries_sum++;
        } /* statbuf */
      } else {
        ret++;
      } /* lstat */
    } else {
      entries_sum++;
    } /* hlinks */
  }
  closedir(dptr);
  if(entries_maxdir!=NULL) {
    if((pathname_maxdir!=NULL)&&(entries_sum>*entries_maxdir)&&(subdir==NULL)) {
      *entries_maxdir=entries_sum;
      *pathname_maxdir='\0';
      str_catn(pathname_maxdir,name,pathname_size);
    }
  }
  *entries+=entries_sum;
  return ret;
}

void putlong(uint32_t num) {
  register uint32_t i,d;
  register int c,o;
  register char *pbuf;
  char buf[14];

  i=num;
  pbuf=buf;
  o=0;
  for(d=1000000000;d>1;d/=10) {
    c=i/d;
    if((c>0)||(o)) {
      *pbuf++=c+'0';
      i=i-(c*d);
      o=1;
    }
  }
  *pbuf++=i+'0';
  *pbuf++='\n';
  write(1,buf,pbuf-buf);
}

void direntries_normal_mode(const char *dir) {
  register uint32_t count;
  register DIR *dptr;
  register struct dirent *eptr;

  dptr=opendir(dir);
  if(dptr==NULL) my_exit(1);
  for(count=0;(eptr=readdir(dptr))!=NULL;count++);
  closedir(dptr);
  putlong(count);
  my_exit(0);
}

void direntries_max_mode(char *basedir, const int32_t depth) {
  register const char *subdir;
  register int ret;
  uint32_t entries;
  uint32_t entries_maxdir;
  char pathname_maxdir[PATHNAME_SIZE+1];
  struct rlimit rl = { 134217728, 134217728 };

  setrlimit(RLIMIT_STACK, &rl);
  ret=str_index(basedir,"/./"); /* "wildcard" like "*" in shell */
  if(ret>0) {
    basedir[ret]='\0';
    subdir=basedir+ret+3;
    ret=walkdir(basedir,&entries,&entries_maxdir,pathname_maxdir,PATHNAME_SIZE,depth,0,subdir);
  } else {
    ret=walkdir(basedir,&entries,&entries_maxdir,pathname_maxdir,PATHNAME_SIZE,depth,0,NULL);
  }
  putlong(entries_maxdir);
  str_catn(pathname_maxdir,"\n",1);
  print(pathname_maxdir);
  my_exit(ret);
}

void direntries_recursive_mode(const char *dir) {
  struct rlimit rl = { 134217728, 134217728 };
  uint32_t entries=0;
  setrlimit(RLIMIT_STACK, &rl);
  int ret=walkdir(dir,&entries,NULL,NULL,0,0,0,NULL);
  putlong(entries);
  my_exit(ret);
}

void direntries_show_mode(register const char *dir, const char *file_extension, const char *file_prefix, register int print_mode) {
  register uint32_t count;
  register DIR *dptr;
  register struct dirent *eptr;
  register char *nptr;
  char pathname[PATHNAME_SIZE+1];
  register uint32_t gifs,jpgs,pngs,jss,csss,phps,htmls,pdfs,exturls;

  dptr=opendir(dir);
  if(dptr==NULL) my_exit(1);
  if(file_extension!=NULL) {
    for(count=0;(eptr=readdir(dptr))!=NULL;) {
      nptr=eptr->d_name;
      if(!str_diff_extension(nptr,file_extension)) {
	count++;
        if(print_mode) {
          *pathname='\0';
          str_catn(pathname,nptr,PATHNAME_SIZE);
          str_catn(pathname,"\n",1);
          print(pathname);
        }
      }
    }
  } else {
    if(file_prefix!=NULL) {
      for(count=0;(eptr=readdir(dptr))!=NULL;) {
        nptr=eptr->d_name;
        if(!str_diffn(nptr,file_prefix,strlen(file_prefix))) {
  	count++;
          if(print_mode) {
            *pathname='\0';
            str_catn(pathname,nptr,PATHNAME_SIZE);
            str_catn(pathname,"\n",1);
            print(pathname);
          }
        }
      }
    } else {
      jpgs=gifs=pngs=jss=csss=phps=htmls=pdfs=exturls=0;
      for(count=0;(eptr=readdir(dptr))!=NULL;count++) {
        nptr=eptr->d_name;
        if(print_mode) {
          *pathname='\0';
          str_catn(pathname,nptr,PATHNAME_SIZE);
          str_catn(pathname,"\n",1);
          print(pathname);
        }
        if(!str_diff_extension(nptr,".gif")) gifs++;
        if(!str_diff_extension(nptr,".jpg")) jpgs++;
        if(!str_diff_extension(nptr,".png")) pngs++;
        if(!str_diff_extension(nptr,".js")) jss++;
        if(!str_diff_extension(nptr,".css")) csss++;
        if(!str_diff_extension(nptr,".php")) phps++;
        if(!str_diff_extension(nptr,".html")) htmls++;
        if(!str_diffn(nptr,"pdf_",4)) pdfs++;
        if(!str_diffn(nptr,"EXTERNAL_URL",12)) exturls++;
      }
      print("*.gif: "); putlong(gifs);
      print("*.jpg: "); putlong(jpgs);
      print("*.png: "); putlong(pngs);
      print("*.js: "); putlong(jss);
      print("*.css: "); putlong(csss);
      print("*.php: "); putlong(phps);
      print("*.html: "); putlong(htmls);
      print("pdf_*: "); putlong(pdfs);
      print("EXTERNAL_URL*: "); putlong(exturls);
    }
  }
  closedir(dptr);
  putlong(count);
  my_exit(0);
}

void direntries_delete_mode(register const char *dir, const char *file_extension, const char *file_prefix) {
  register uint32_t count;
  register DIR *dptr;
  register struct dirent *eptr;
  register char *nptr;
  char pathname[3*PATHNAME_SIZE];

  dptr=opendir(dir);
  if(dptr==NULL) my_exit(1);
  if(file_extension!=NULL) {
    for(count=0;(eptr=readdir(dptr))!=NULL;) {
      nptr=eptr->d_name;
      if(!str_diff_extension(nptr,file_extension)) {
        count++;
        nptr=eptr->d_name;
	/*
        str_catn(pathname,"deleting: ",PATHNAME_SIZE);
        str_catn(pathname,"\n",1);
	print(pathname);
	*/
        *pathname='\0';
        str_catn(pathname,dir,PATHNAME_SIZE);
        str_catn(pathname,"/",1);
        str_catn(pathname,nptr,PATHNAME_SIZE);
        unlink(pathname);
      }
    }
  } else {
    if(file_prefix!=NULL) {
      for(count=0;(eptr=readdir(dptr))!=NULL;) {
        nptr=eptr->d_name;
        if(!str_diffn(nptr,file_prefix,strlen(file_prefix))) {
          count++;
          nptr=eptr->d_name;
  	/*
          str_catn(pathname,"deleting: ",PATHNAME_SIZE);
          str_catn(pathname,"\n",1);
  	print(pathname);
  	*/
          *pathname='\0';
          str_catn(pathname,dir,PATHNAME_SIZE);
          str_catn(pathname,"/",1);
          str_catn(pathname,nptr,PATHNAME_SIZE);
          unlink(pathname);
        }
      }
    } else {
      for(count=0;(eptr=readdir(dptr))!=NULL;count++) {
        nptr=eptr->d_name;
        *pathname='\0';
        str_catn(pathname,dir,PATHNAME_SIZE);
        str_catn(pathname,"/",1);
        str_catn(pathname,nptr,PATHNAME_SIZE);
        unlink(pathname);
      }
    }
  }
  closedir(dptr);
  putlong(count);
  my_exit(0);
}

int main(int argc,char *const argv[]) {
  register int c;
  register char *dir;
  register char *file_extension;
  register char *file_prefix;
  int32_t depth;
  register int mode;
  register int print_mode;
  const char wd[]=".";

  file_prefix=file_extension=NULL;
  dir=(char *)wd;
  print_mode=0;
  depth=2;
  mode=MODE_NORMAL;
  for(c=1;c<argc;c++) {
    if(str_diffn(argv[c],"-h",3)==0) { print(usage); _exit(0); }
    if(str_diffn(argv[c],"-r",3)==0) {
      mode=MODE_RECURSIVE;
      continue;
    }
    if(str_diffn(argv[c],"-m",3)==0) {
      mode=MODE_MAXIMUM;
      continue;
    }
    if(str_diffn(argv[c],"-s",3)==0) {
      mode=MODE_SHOW;
      continue;
    }
    if(str_diffn(argv[c],"-D",3)==0) {
      mode=MODE_DELETE;
      continue;
    }
    if(str_diffn(argv[c],"-p",3)==0) {
      print_mode=1;
      continue;
    }
    if(str_diffn(argv[c],"-d",3)==0) {
      c++;
      depth=(uint32_t)str2uint63(argv[c]);
      continue;
    }
    if(str_diffn(argv[c],"-e",3)==0) {
      c++;
      file_extension=argv[c];
      continue;
    }
    if(str_diffn(argv[c],"-P",3)==0) {
      c++;
      file_prefix=argv[c];
      continue;
    }
    if(str_diffn(argv[c],"-",1)!=0) {
      dir=argv[c];
    }
  }
  if(print_mode&&(mode!=MODE_SHOW)) { print(usage); _exit(1); }
  switch(mode) {
    case MODE_NORMAL:
      direntries_normal_mode(dir);
      break;
    case MODE_RECURSIVE:
      direntries_recursive_mode(dir);
      break;
    case MODE_MAXIMUM:
      direntries_max_mode(dir,depth);
      break;
    case MODE_SHOW:
      direntries_show_mode(dir,file_extension,file_prefix,print_mode);
      break;
    case MODE_DELETE:
      direntries_delete_mode(dir,file_extension,file_prefix);
      break;
    default:
      print(usage);
      _exit(1);
      break;
  }
  _exit(0);
}
