/**
 * filescan.c
 *
 * Compile with:
 * diet gcc -O2 -Wall -o filescan filescan.c && sstrip filescan
 * Frank Bergmann, www.tuxad.com, 11/2008
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>
#include <unistd.h>
#include <time.h>
#include <stdint.h>

#include <string.h>
#undef USE_STRING_LIB
#define USE_STRING_LIB


#define MAX_CHILDS 20
#define SLEEP 5

extern char **environ;
static const char msg[]="usage: filescan [-d changedir] [-I includestring] [-X excludestring] path/to/command [arg(s)]\n";


#ifndef USE_STRING_LIB

int strlen(register const char* in) {
  register const char *t = in;
  if (t == NULL) return 0;
  for (;;) { if (!*t) break; ++t; }
  return (int)(t-in);
}

int strncmp(register const char* s, register const char* t, int limit) {
  register const char* u=t+limit;
  register int j;
  j=0;
  for (;;) { if (t>=u) break; if ((j=(*s-*t))) break; if (!*t) break; ++s; ++t; }
  return j;
}

char *strcpy(register char* out, register const char* in) {
  for (;;) { if (!*in) break; *out++=*in++; }
  *out='\0';
  return out;
}

#endif

int32_t str_index(const char* a, const char* b) {
  register const char* s=a;
  register int t=strlen(b);
  register const char* u=s+strlen(s)-t;
  register char z;

  z = *b;
  for (;;) { if (s>u) break; if ((*s==z) && (!strncmp(s,b,t))) return s-a; ++s; }
  return -1;
}

void my_exit(int r,const char *s) {
  write(2,s,strlen(s));
  _exit(r);
}

int main(int argc,char *argv[]) {
  register char *p;
  register int c;
  register int firstfree;
  register char *exclude_string;
  register char *include_string;
  int childstatus;
  int sstring;
  char *changedir;
  char fnames[MAX_CHILDS][256];
  pid_t cpids[MAX_CHILDS];
  time_t ctimes[MAX_CHILDS];
  pid_t child;
  struct dirent *dent;
  DIR *curdir;
  char **args;

  if (argc < 2) my_exit(127, msg);
  changedir = 0;
  sstring = -1;
  exclude_string = include_string = 0;
  for (c = 1; (c < argc) && (sstring == -1); c++) {
    if (*argv[c] == '-') {
      if (argv[c][2] != '\0') my_exit(127, msg);
      switch (argv[c][1]) {
        case 'd': changedir = argv[++c]; break;
        case 'I': include_string = argv[++c]; break;
        case 'X': exclude_string = argv[++c]; break;
        default: my_exit(127, msg);
      }
    } else {
      sstring = c;
      break;
    }
  }
  if (sstring == -1) my_exit(127, msg);

  if (changedir && (chdir(changedir) != 0)) my_exit(111, "can't change dir\n");

  sstring--;
  args = &argv[sstring];
  /* move array elements by 1 back to add filename for execve() */
  for (c=sstring;c<argc-1;c++) {
    argv[c] = argv[c+1];
  }

  for(c=0;c<MAX_CHILDS;c++) fnames[c][0]=0;


  /* main loop */
  for(;;) {

    curdir = opendir(".");
    for (;;) {
  
      next_dirent:
      dent = readdir(curdir);
      if (!dent) break;

      /* does file match any rules? */
      p = dent->d_name;
      if (*p == '.') continue; /* skip names beginning with dot */
      if (exclude_string && (str_index(p, exclude_string) != -1)) continue;
      if (include_string && (str_index(p, include_string) == -1)) continue;

      /* file already processed? */
      firstfree = -1;
      for (c=0;c<MAX_CHILDS;c++) {
        if (*fnames[c]) {
          if (!strncmp(fnames[c], p, strlen(p))) {
            /* file is processed by a child, next file */
            goto next_dirent;
          }
        } else {
          if (firstfree == -1) {
            /* first free array element */
            firstfree = c;
          }
        }
      }

      /* array full? => sleep and run again */
      c = firstfree;
      if (c == -1) goto closedir_and_sleep;

      /* file's not already processed and array's not full */
      ctimes[c] = time(0);
      strcpy(fnames[c], p);
      argv[argc-1] = fnames[c];

      /* we just use execve() after forking, use vfork() */
      child = vfork();
      if (child == -1) {
        /* error on vfork() */
        *fnames[c] = 0;
        write(2, "can't fork\n", 11);
        goto closedir_and_sleep;
      }

      if (child) {
        /* father */
        cpids[c] = child; /* assigned after execve() */
      } else {
        /* child */
        execve(args[0], args, environ);
        *fnames[c] = 0;
        write(2,"exec err\n",9);
        _exit(111);
      }
  
    }
  
    closedir_and_sleep:
    closedir(curdir);

    /* delete array entries of terminated childs */
    for (c=0;c<MAX_CHILDS;c++) {
      if ((*fnames[c]) || (cpids[c] != 0)) {
        if (waitpid(cpids[c], &childstatus, WNOHANG) == cpids[c]) {
          if (WIFEXITED(childstatus) || WIFSIGNALED(childstatus)) {
            *fnames[c] = 0;
            cpids[c] = 0;
          }
        }
      }
    }

    sleep(SLEEP);
  }

  /* never reached */
  _exit(0);
}

