/**
 * sfgrep.c
 * simple-fixed-grep or string-fast-grep
 * works similar to grep but with strings not patterns
 * Compile with:
 * diet -Os gcc -Os -Wall -o sfgrep sfgrep.c && sstrip sfgrep
 * Frank Bergmann, www.tuxad.com, 08/2008
 */

#define _FILE_OFFSET_BITS 64

#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdint.h>

#define READBUFSIZE 1048576
#define WRITEBUFSIZE 8192

static const char msg[]="missing arg(s)\nusage: sfgrep [-h] [-H] [-v] [-c] [-l] [-i] string [file(s)]\n";
static const char warning_open[]=" * warning, could not be opened\n";
static const char warning_read[]=" * warning, could not be read\n";

typedef struct buffer {
  char *s;
  char *c;
  char *t;
  int fd;
} buffer;

typedef struct grep_parms {
  char *search_string;
  buffer *write_buffer;
  char *filename;
  int fnlen;
  int mode_withfilename;
  int mode_firstmatch;
  int mode_invert;
  int mode_insenscase;
  int mode_count;
  uint32_t matchcount;
} grep_parms;

uint32_t str_len(register const char* in) {
  register const char* t=in;
  for (;;) {
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
  }
  return (uint32_t)(t-in);
}

int buffer_flush(buffer *b) {
  if (b->c == b->s) return 0;
  return write(b->fd, b->s, b->c - b->s);
}

int buffer_putc(buffer *b, register const char a) {
  register char *p;
  register char *s;

  p = b->c;
  if (p < b->t) goto put;

  s = b->s;
  if (write(b->fd, s, p - s) < 0) return -1;
  p = s;

  put:
  *p++ = a;
  b->c = p;
  return 0;
}

int buffer_put(buffer *b, register const char *a, register uint32_t l) {
  register char *p;
  register char *s;
  register const char *t;

  if (!l) return 0;
  p = b->c;
  t = b->t;
  if ((p+l) <= t) goto put;

  s = b->s;
  if (write(b->fd, s, p-s) < 0) return -1;
  p = s;
  if ((p+l) > t) {
    b->c = p;
    return write(b->fd, a, l);
  }

  put:
  t = a + l;
  for(;;) {
    *p++ = *a++; if (a >= t) break;
    *p++ = *a++; if (a >= t) break;
    *p++ = *a++; if (a >= t) break;
    *p++ = *a++; if (a >= t) break;
  }
  b->c = p;
  return 0;
}

void buffer_putlongnl(buffer *b, uint32_t num) {
  register uint32_t i,d;
  register int c,o;
  register char *pbuf;
  char buf[14];

  i=num; o=0;
  pbuf=buf;
  for(d=1000000000;d>1;d/=10) {
    c=i/d;
    if((c>0)||(o)) {
      *pbuf++=c+'0';
      i=i-(c*d);
      o=1;
    }
  }
  *pbuf++=i+'0'; *pbuf++='\n';
  buffer_put(b, buf, pbuf-buf);
}

void my_exit(int retval, const char *message) {
  if (message && (*message != '\0')) write(2, message, str_len(message));
  _exit(retval);
}


uint32_t grep_filebuffer(register char *p, register char *t, grep_parms *grep) {
  register char *line;
  register char *s;
  register char *search_string;
  register int invert;
  register int mode_withfilename;
  buffer *write_buffer;

  write_buffer = grep->write_buffer;
  invert = grep->mode_invert;
  mode_withfilename = grep->mode_withfilename;
  search_string = grep->search_string;
  s = search_string;

  for(;;) {

    next_line:
    line = p;

    for(;;) {

      if (*p == '\n') break;
      if (p >= t) return 0;

      string_cmp:
      if ((*p == *s) || (grep->mode_insenscase && (*p >= 'A') && (*p <= 'Z') && (((*p)+32) == *s))) {
        next_char:
        ++p; ++s;
        if (*s == '\0') {
          /* match */
          s = search_string;
          for(;;) {
            /* don't use: "if (*p == '\n') break; p++;" */
            if (p >= t) break; if (*p++ == '\n') break;
            if (p >= t) break; if (*p++ == '\n') break;
            if (p >= t) break; if (*p++ == '\n') break;
            if (p >= t) break; if (*p++ == '\n') break;
          }
          if (!invert) {
            grep->matchcount += 1;
            if (mode_withfilename && !grep->mode_count) {
              buffer_put(write_buffer, grep->filename, grep->fnlen);
              if (grep->mode_firstmatch) {
                buffer_putc(write_buffer, '\n');
                return 1;
              }
              buffer_putc(write_buffer, ':');
            }
            if (grep->mode_firstmatch) return 1;
            if (!grep->mode_count) buffer_put(write_buffer, line, p-line);
          }
          goto next_line;
        } /* *s == '\0' */
        if (*p == '\n') { s = search_string; break; }
        if (p >= t) return 0;
        if ((*p == *s) || (grep->mode_insenscase && (*p >= 'A') && (*p <= 'Z') && (((*p)+32) == *s))) goto next_char;
        s = search_string;
        goto string_cmp;
      } /* *p == *s */

      ++p;
    }

    for(;;) {
      ++p;
      if (invert) {
        grep->matchcount += 1;
        if (mode_withfilename && !grep->mode_count) {
          buffer_put(write_buffer, grep->filename, grep->fnlen);
          if (grep->mode_firstmatch) {
            buffer_putc(write_buffer, '\n');
            return 1;
          }
          buffer_putc(write_buffer, ':');
        }
        if (grep->mode_firstmatch) return 1;
        if (!grep->mode_count) buffer_put(write_buffer, line, p-line);
        line = p;
      }
      if (p >= t) return 0;
      if (*p != '\n') break;
    }

  }
}

void grep_file(int fd, grep_parms *grep) {
  register char *p;
  register char *t;
  register char *n;
  char rbufspace[2*READBUFSIZE+1];
  register ssize_t rb;

  p = rbufspace;
  grep->matchcount = 0;
  for(;;) {
    rb = read(fd, p, READBUFSIZE);
    if (rb < 0) {
        if (grep->filename) write(2,grep->filename,grep->fnlen);
        write(2,warning_read,str_len(warning_read));
        return;
    }
    if (rb == 0) {
      if (grep->mode_count) {
        if (grep->mode_withfilename) {
          buffer_put(grep->write_buffer, grep->filename, grep->fnlen);
          buffer_putc(grep->write_buffer, ':');
        }
        buffer_putlongnl(grep->write_buffer, grep->matchcount);
      }
      return;
    }
    t = p + rb;
    p = rbufspace;
    /* find last \n */
    n = t;
    for(;;) {
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
      --n; if (*n == '\n') break; if (n <= p) break;
    }
    if (n <= p) {
      n = t;
    } else {
      ++n;
    }
    if (grep_filebuffer(p, n, grep) && grep->mode_firstmatch) return;
    if (n != t) {
      for(;;) {
        *p++ = *n++; if (n >= t) break;
        *p++ = *n++; if (n >= t) break;
        *p++ = *n++; if (n >= t) break;
        *p++ = *n++; if (n >= t) break;
      }
    }
  }
}


int main(int argc,char *const argv[]) {
  register char *p;
  register int c;
  char wbufspace[WRITEBUFSIZE+1];
  int fd;
  int file1;
  int sstring;
  int optsend;
  int retval;
  grep_parms grep;

  if (argc < 2) my_exit(127, msg);
  optsend = 0;
  retval = 1;
  grep.mode_insenscase = 
    grep.mode_invert = 
    grep.mode_firstmatch = 
    grep.mode_count = 0;
  sstring = file1 = grep.mode_withfilename = -1;
  for(c=1;c<argc;c++) {
    if (!optsend && (*argv[c] == '-')) {
      if (argv[c][2] != '\0') my_exit(127, msg);
      switch (argv[c][1]) {
        case 'v': grep.mode_invert = 1; break;
        case 'h': grep.mode_withfilename = 0; break;
        case 'H': grep.mode_withfilename = 1; break;
        case 'c': grep.mode_count = 1; break;
        case 'l': grep.mode_firstmatch = 1; break;
        case 'i': grep.mode_insenscase = 1; break;
        case '-': optsend = 1; break;
        default: break;
      }
    } else {
      if (sstring == -1) {
        sstring = c;
      } else {
        if (file1 == -1) {
          file1 = c;
        } else {
          if (grep.mode_withfilename == -1) grep.mode_withfilename = 1;
        }
      }
    }
  }
  if (sstring == -1) my_exit(127, msg);
  if (grep.mode_count && (grep.mode_withfilename == -1)) grep.mode_withfilename = 1;
  if (grep.mode_withfilename == -1) grep.mode_withfilename = 0;
  if (grep.mode_firstmatch) grep.mode_withfilename = 1;
  /* F*cking illogical GNU-grep compatibility fix: */
  if (grep.mode_firstmatch) grep.mode_count = 0;

  p = wbufspace;
  buffer wbuf = { p, p, p+WRITEBUFSIZE, 1 };
  grep.write_buffer = &wbuf;
  grep.search_string = argv[sstring];
  if (grep.mode_insenscase) {
    p = grep.search_string;
    for(;;) {
      if (!*p) break; if ((*p >= 'A') && (*p <= 'Z')) *p += 32; ++p;
    }
  }

  if (file1 == -1) {
    grep.filename = NULL;
    grep.fnlen = 0;
    grep.mode_withfilename = 0;
    grep_file(0, &grep);
  } else {
    for(c=file1;c<argc;c++) {
      fd = open(argv[c], O_RDONLY);
      if (fd < 0) {
        write(2,argv[c],str_len(argv[c]));
        write(2,warning_open,str_len(warning_open));
      } else {
        grep.filename = argv[c];
        grep.fnlen = str_len(grep.filename);
        grep_file(fd, &grep);
        if (grep.matchcount) retval = 0;
        close(fd);
      }
    }
  }
  buffer_flush(&wbuf);
  if (grep.matchcount) retval = 0;
  _exit(retval);
}

